@@ -173,12 +173,13 @@ linestrings_from_graph <- function(graph_df, crs = 4326) {
 #'
 #' @export
 #' @importFrom collapse ftransform GRP get_vars add_vars add_vars<- ffirst colorderv %!in% collap
+#' @importFrom kit fpmin fpmax
 create_undirected_graph <- function(graph_df, by = NULL, ...) {
   if(length(by)) {
     if(is.call(by)) by <- all.vars(by)
     if(!is.character(by)) stop("by needs to be a one-sided formula or a character vector of column names")
   }
-  graph_df <- ftransform(graph_df, from = pmin(from, to), to = pmax(from, to))
+  graph_df <- ftransform(graph_df, from = fpmin(from, to), to = fpmax(from, to))
   g <- GRP(graph_df, c("from", "to", by), sort = FALSE)
   nam <- names(graph_df)
   agg_first <- c("edge", "FX", "FY", "TX", "TY")
@@ -250,7 +251,7 @@ nodes_from_graph <- function(graph_df, sf = FALSE, crs = 4326) {
 }
 
 #' @title Compute Distance Matrix from Graph
-#' @description Compute a distance matrix for all node pairs in a graph using cppRouting.
+#' @description Compute a distance matrix for all node pairs in a graph using \emph{igraph}.
 #'
 #' @param graph_df A data frame representing a graph with columns:
 #'   \code{from}, \code{to}, and \code{cost}.
@@ -267,7 +268,7 @@ nodes_from_graph <- function(graph_df, sf = FALSE, crs = 4326) {
 #' @details
 #' This function:
 #' \itemize{
-#'   \item Converts the graph data frame to a cppRouting graph object
+#'   \item Converts the graph data frame to a \emph{igraph} graph object
 #'   \item Contracts the graph for efficient distance computation
 #'   \item Computes the distance matrix for all node pairs using the specified algorithm
 #' }
@@ -384,22 +385,20 @@ normalize_graph <- function(graph_df) {
 #' @param keep.nodes Numeric vector (optional). Node IDs to preserve during consolidation,
 #'   even if they occur exactly twice. Also used to preserve nodes when dropping singleton edges.
 #' @param \dots Arguments passed to \code{\link[collapse]{collap}()} for aggregation across contracted edges. The defaults are \code{FUN = fmean} for numeric columns and \code{catFUN = fmode} for categorical columns. Select columns using \code{cols} or use argument \code{custom = list(fmean = cols1, fsum = cols2, fmode = cols3)} to map different columns to specific aggregation functions. It is highly recommended to weight the aggregation (using \code{w = ~ weight_col}) by the length/cost of the edges.
-#' @param recursive One of \code{"none"/FALSE}, \code{"partial"} (recurse on dropping single edges and consolidation but only aggregate once), or \code{"full"/TRUE} (recursively contracts and aggregates the graph
-#'   until no further consolidation is possible). This ensures that long chains of intermediate
-#'   nodes are fully contracted in a single call.
+#' @param recursive One of \code{"none"/FALSE} (drop edges, contract, and aggregate once), \code{"partial"} (recursively drop edges and contract but only aggregate once), or \code{"full"/TRUE} (recursively drop edges, contract, and aggregate the graph
+#'   until no further consolidation is possible).
 #' @param verbose Logical (default: TRUE). Whether to print messages about dropped edges
 #'   and consolidation progress.
 #'
 #' @return A data frame representing the consolidated graph with:
 #'   \itemize{
-#'     \item \code{edge} - Edge identifier (added as first column)
 #'     \item All columns from \code{graph_df} (aggregated if consolidation occurred),
-#'       excluding \code{from}, \code{to}, and optionally \code{FX}, \code{FY}, \code{TX}, \code{TY}
+#'       excluding \code{from}, \code{to}, and optionally \code{edge} and \code{FX}, \code{FY}, \code{TX}, \code{TY}
 #'       (which are re-added if present in original)
-#'     \item \code{from}, \code{to} - Node IDs (updated after consolidation)
+#'     \item \code{from}, \code{to}, \code{edge} - Node/edge IDs (updated after consolidation)
 #'     \item Coordinate columns (\code{FX}, \code{FY}, \code{TX}, \code{TY}) if present in original
-#'     \item Attribute \code{"keep.edges"} - Indices of original edges that were kept
-#'     \item Attribute \code{"gid"} - Edge group IDs mapping contracted edges to original edges
+#'     \item Attribute \code{"keep.edges"} - Indices of original edges that were kept (before aggregation)
+#'     \item Attribute \code{"group.id"} - Integer mapping each kept edge to its row in the result (after aggregation)
 #'   }
 #'
 #' @details
@@ -449,7 +448,7 @@ normalize_graph <- function(graph_df) {
 #' c(original = nrow(graph), consolidated = nrow(graph_cons))
 #'
 #' @export
-#' @importFrom collapse fnrow get_vars anyv setv ss seq_row fduplicated fmatch whichv whichNA allNA ffirst GRP collap %iin% %!in% %!iin% join colorderv funique.default %!=% %==% missing_cases qtab flast varying radixorderv groupv na_rm
+#' @importFrom collapse fnrow get_vars anyv setv ss seq_row fduplicated fmatch whichv whichNA allNA ffirst GRP collap %iin% %!in% %!iin% join colorderv funique.default %!=% %==% missing_cases qtab varying radixorderv groupv na_rm
 #' @importFrom kit countOccur
 #' @importFrom stats setNames
 consolidate_graph <- function(graph_df, directed = FALSE,
@@ -461,7 +460,10 @@ consolidate_graph <- function(graph_df, directed = FALSE,
 
   if(...length() && any(...names() == "consolidate")) contract <- list(...)[["consolidate"]]
 
-  if(verbose) namg <- flast(as.character(substitute(graph_df)))
+  if(verbose) {
+    namg <- as.character(substitute(graph_df))
+    if(length(namg) != 1L) namg <- "."
+  }
 
   reci <- switch(as.character(recursive), none =, `FALSE` = 0L, partial = 1L, full =, `TRUE` = 2L,
                  stop("recursive needs to be one of 'none'/FALSE, 'partial', or 'full'/TRUE"))
@@ -508,13 +510,14 @@ consolidate_graph <- function(graph_df, directed = FALSE,
                                     keep.nodes = keep.nodes,
                                     reci = reci, nam_keep = nam_keep,
                                     verbose = verbose, ...)
+      if(length(attr(res, ".early.return"))) break
     }
   }
 
   if(length(attr(res, ".early.return"))) attr(res, ".early.return") <- NULL
 
   if(verbose) {
-    cat(sprintf("\nConsolidated %s graph %s from %d edges to %d edges (%s%%)\n", if(directed) "directed" else "undirected", namg, fnrow(graph_df), fnrow(res), as.character(signif(100*fnrow(res)/fnrow(graph_df), 3))))
+    cat(sprintf("\nConsolidated %s graph %s from %d edges to %d edges (%.1f%%)\n", if(directed) "directed" else "undirected", namg, fnrow(graph_df), fnrow(res), 100*fnrow(res)/fnrow(graph_df)))
     print(qtab(countOccur(c(res$from, res$to))$Count, dnn = "Final node degrees:"))
   }
 
@@ -523,7 +526,7 @@ consolidate_graph <- function(graph_df, directed = FALSE,
     if(any(nam_rm[3:4] %in% nam)) res <- join(res, setNames(nodes, c("from", "FX", "FY")), on = "from", verbose = 0L)
     if(any(nam_rm[5:6] %in% nam)) res <- join(res, setNames(nodes, c("to", "TX", "TY")), on = "to", verbose = 0L)
   }
-  add_vars(res, pos = "front") <- list(edge = seq_row(res))
+  if(anyv(nam, "edge")) add_vars(res, pos = "front") <- list(edge = seq_row(res))
 
   # Reordering columns
   res <- colorderv(res, radixorderv(fmatch(names(res), nam)))
@@ -660,7 +663,7 @@ consolidate_graph_core <- function(graph_df, directed = FALSE,
     degree_table <- compute_degrees(gft$from, gft$to)
     if(!fnrow(degree_table)) break
 
-    if(anyv(drop.edges, "single") && anyv(degree_table$deg_total, 1L)) {
+    if(anyv(drop.edges, "single") && reci > 0L && anyv(degree_table$deg_total, 1L)) {
       nodes <- degree_table$node[degree_table$deg_total %==% 1L]
       if(length(keep.nodes)) nodes <- nodes[nodes %!iin% keep.nodes]
       if(length(nodes)) {
@@ -763,7 +766,7 @@ compute_degrees <- function(from_vec, to_vec) {
 #'   representing node coordinates.
 #' @param nodes For \code{method = "shortest-paths"}: either an atomic vector of node IDs, or a
 #'   data.frame with columns \code{from} and \code{to} specifying origin-destination pairs.
-#'   For \code{method = "cluster"}: an atomic vector of node IDs to preserve. These nodes will
+#'   For \code{method = "cluster"}: an (optional) atomic vector of node IDs to preserve. These nodes will
 #'   be kept as cluster centroids, and nearby nodes (within \code{radius_km$nodes}) will be
 #'   assigned to their clusters. Remaining nodes are clustered using \code{\link[leaderCluster]{leaderCluster}}.
 #' @param method Character string (default: "shortest-paths"). Method to use for simplification:
@@ -788,6 +791,7 @@ compute_degrees <- function(from_vec, to_vec) {
 #'   \code{nodes}: radius in kilometers around preserved nodes. Graph nodes within this radius
 #'   will be assigned to the nearest preserved node's cluster.
 #'   \code{cluster}: radius in kilometers for clustering remaining nodes using leaderCluster.
+#' @param verbose Logical (default: TRUE). Whether to print progress messages/bars.
 #' @param \dots For \code{method = "cluster"}: additional arguments passed to
 #'   \code{\link[collapse]{collap}} for edge attribute aggregation.
 #'
@@ -881,14 +885,14 @@ compute_degrees <- function(from_vec, to_vec) {
 #' }
 #'
 #' @export
-#' @importFrom collapse fnrow ss ckmatch funique.default fmatch gsplit fmin dapply whichv %+=% GRP add_vars seq_row add_stub colorderv %!in% collap get_vars alloc
+#' @importFrom collapse fnrow ss ckmatch funique funique.default fmatch gsplit fmin dapply whichv %+=% GRP add_vars seq_row add_stub colorderv %!in% collap get_vars alloc
 #' @importFrom kit iif
 #' @importFrom igraph graph_from_data_frame delete_vertex_attr igraph_options shortest_paths
 #' @importFrom geodist geodist_vec geodist_min
 #' @importFrom leaderCluster leaderCluster
-simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "cluster"),
+simplify_network <- function(graph_df, nodes = NULL, method = c("shortest-paths", "cluster"),
                              directed = FALSE, cost.column = "cost", by = NULL,
-                             radius_km = list(nodes = 7, cluster = 20), ...) {
+                             radius_km = list(nodes = 7, cluster = 20), verbose = TRUE, ...) {
 
   method <- match.arg(method)
 
@@ -925,6 +929,8 @@ simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "clus
                             vertices = data.frame(name = seq_along(all_nodes))) |>
       delete_vertex_attr("name")
 
+    if(verbose) cat("Created graph with", vcount(ig), "nodes and", ecount(ig), "edges...\n")
+
     # Don't return vertex/edge names
     iopt <- igraph_options(return.vs.es = FALSE)
     on.exit(igraph_options(iopt))
@@ -933,11 +939,12 @@ simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "clus
 
     # Pre-process nodes input ONCE (before any group loop)
     if(is.atomic(nodes)) {
-      nodes_matched <- ckmatch(nodes, all_nodes, e = "Unknown nodes:")
+      nodes_matched <- ckmatch(funique.default(nodes), all_nodes, e = "Unknown nodes:")
       nodes_split <- NULL
     } else if(is.data.frame(nodes)) {
       if(!all(c("from", "to") %in% names(nodes)))
         stop("nodes data.frame must have columns 'from' and 'to'")
+      nodes <- funique(nodes, cols = c("from", "to"))
       nodes_split <- gsplit(ckmatch(nodes$to, all_nodes, e = "Unknown 'to' nodes:"),
                             ckmatch(nodes$from, all_nodes, e = "Unknown 'from' nodes:"),
                             use.g.names = TRUE)
@@ -947,24 +954,41 @@ simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "clus
     # Helper to compute paths with given cost vector
     compute_paths <- function(cost_vec) {
       if(length(nodes_matched)) {
+        N <- length(nodes_matched)
         if(directed) {
+          if(verbose) {
+            pb <- progress_bar$new(format = "Processed :current/:total shortest paths (:percent) at :tick_rate/sec [Elapsed::elapsed | ETA::eta]",
+                                   total = N^2, clear = FALSE, width = 100)
+          }
           for (i in nodes_matched) {
+            if(verbose) pb$tick(N)
             pathsi <- shortest_paths(ig, from = i, to = nodes_matched,
                                      weights = cost_vec, mode = "out", output = "epath")$epath
             .Call(C_mark_edges_traversed, pathsi, edges_traversed)
           }
         } else {
-          n <- length(nodes_matched)
-          for (i in 1:(n - 1L)) {
-            pathsi <- shortest_paths(ig, from = nodes_matched[i], to = nodes_matched[(i + 1L):n],
+          if(verbose) {
+            pb <- progress_bar$new(format = "Processed :current/:total shortest paths (:percent) at :tick_rate/sec [Elapsed::elapsed | ETA::eta]",
+                                   total = N*(N-1)/2, clear = FALSE, width = 100)
+          }
+          for (i in 1:(N - 1L)) {
+            ind = nodes_matched[(i + 1L):N]
+            if(verbose) pb$tick(length(ind))
+            pathsi <- shortest_paths(ig, from = nodes_matched[i], to = ind,
                                      weights = cost_vec, mode = "out", output = "epath")$epath
             .Call(C_mark_edges_traversed, pathsi, edges_traversed)
           }
         }
       } else {
         from_nodes <- as.integer(names(nodes_split))
+        if(verbose) {
+          pb <- progress_bar$new(format = "Processed :current/:total shortest paths (:percent) at :tick_rate/sec [Elapsed::elapsed | ETA::eta]",
+                                 total = fnrow(nodes), clear = FALSE, width = 100)
+        }
         for (i in seq_along(from_nodes)) {
-          pathsi <- shortest_paths(ig, from = from_nodes[i], to = nodes_split[[i]],
+          ind = nodes_split[[i]]
+          if(verbose) pb$tick(length(ind))
+          pathsi <- shortest_paths(ig, from = from_nodes[i], to = ind,
                                    weights = cost_vec, mode = "out", output = "epath")$epath
           .Call(C_mark_edges_traversed, pathsi, edges_traversed)
         }
@@ -986,6 +1010,9 @@ simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "clus
 
     edges <- which(edges_traversed > 0L)
     result <- ss(graph_df, edges, check = FALSE)
+    if(verbose) cat(sprintf("Retained %d/%d edges traversed by shortest paths (%.1f%%)\n",
+                            length(edges), fnrow(graph_df),
+                            100 * length(edges) / fnrow(graph_df)))
     attr(result, "edges") <- edges
     attr(result, "edge_counts") <- edges_traversed[edges]
 
@@ -999,10 +1026,11 @@ simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "clus
     # Optional node weights
     if(is.numeric(cost.column) && length(cost.column) == fnrow(nodes_df)) nodes_df$weights <- cost.column
     # Cluster method
-    cl <- cluster_nodes(nodes_df, nodes, radius_km$nodes, radius_km$cluster)
+    cl <- cluster_nodes(nodes_df, funique.default(nodes), radius_km$nodes, radius_km$cluster, verbose)
     # Graph Contraction to Clusters
     result <- contract_edges(graph_df, nodes = nodes_df, clusters = cl$clusters,
-                             centroids = cl$centroids, directed = directed, by = by, ...)
+                             centroids = cl$centroids, directed = directed, by = by,
+                             verbose = verbose, ...)
 
   }
 
@@ -1012,12 +1040,13 @@ simplify_network <- function(graph_df, nodes, method = c("shortest-paths", "clus
 # Helper functions for simplify_network() with method = "cluster"
 cluster_nodes <- function(nodes, keep,
                           nodes_radius_km = 7,
-                          cluster_radius_km = 20) {
+                          cluster_radius_km = 20, verbose = TRUE) {
   # Nodes to preserve
   if(length(keep)) {
     clusters <- integer(fnrow(nodes))
     keep <- ckmatch(keep, nodes$node, "Unknown nodes to preserve:")
     clusters[keep] <- seq_along(keep)
+    if(verbose) cat("Clustering nodes close to 'keep' nodes using a radius of ", nodes_radius_km, "km\n", sep = "")
     # Cluster nodes close to cities
     dmat <- geodist_vec(nodes$X[keep], nodes$Y[keep],
                         nodes$X[-keep], nodes$Y[-keep],
@@ -1030,7 +1059,10 @@ cluster_nodes <- function(nodes, keep,
     if(length(ind)) {
       mat <- cbind(Y = nodes$Y[ind], X = nodes$X[ind])
       weights <- if(length(nodes$weights)) nodes$weights[ind] else alloc(1, nrow(mat))
-      res <- leaderCluster(mat, cluster_radius_km, weights, max_iter = 150L, distance = "haversine")
+      if(verbose) cat("Clustering the remaining nodes with the leaderCluster algorithm using a radius of ", cluster_radius_km, "km\n", sep = "")
+      res <- leaderCluster(mat, cluster_radius_km, weights, max_iter = 250L, distance = "haversine")
+      if(res$iter >= 250) warning("leaderCluster did not fully converge within 250 iterations")
+      else if(verbose) cat("leaderCluster algorithm converged in", res$iter, "iterations\n")
       clusters[ind] <- res$cluster_id %+=% length(keep)
       centroids <- integer(length(keep) + res$num_clusters)
       centroids[seq_along(keep)] <- keep
@@ -1039,7 +1071,9 @@ cluster_nodes <- function(nodes, keep,
   } else {
     mat <- cbind(Y = nodes$Y, X = nodes$X)
     weights <- if(length(nodes$weights)) nodes$weights[ind] else alloc(1, nrow(mat))
-    res <- leaderCluster(mat, cluster_radius_km, weights, max_iter = 150L, distance = "haversine")
+    res <- leaderCluster(mat, cluster_radius_km, weights, max_iter = 250L, distance = "haversine")
+    if(res$iter >= 250) warning("leaderCluster did not fully converge within 250 iterations")
+    else if(verbose) cat("leaderCluster algorithm converged in", res$iter, "iterations\n")
     clusters <- res$cluster_id
     centroids <- res$cluster_centroids[,2:1]
     dimnames(centroids)[[2L]] <- c("X", "Y")
@@ -1048,7 +1082,7 @@ cluster_nodes <- function(nodes, keep,
   list(clusters = clusters, centroids = nodes$node[centroids])
 }
 
-contract_edges <- function(graph, nodes, clusters, centroids, directed = FALSE, by = NULL, ...) {
+contract_edges <- function(graph, nodes, clusters, centroids, directed = FALSE, by = NULL, verbose = TRUE, ...) {
     node_centroids <- centroids[clusters]
     graph$from <- node_centroids[ckmatch(graph$from, nodes$node)]
     graph$to <- node_centroids[ckmatch(graph$to, nodes$node)]
@@ -1056,6 +1090,7 @@ contract_edges <- function(graph, nodes, clusters, centroids, directed = FALSE,
     # Drop self-loops (edges where both endpoints map to same cluster)
     self_loops <- graph$from == graph$to
     if(any(self_loops)) {
+      if(verbose) cat("Dropped", sum(self_loops), "self-loop edges (following clustering)\n")
       graph <- ss(graph, !self_loops, check = FALSE)
     }
 
@@ -1063,6 +1098,7 @@ contract_edges <- function(graph, nodes, clusters, centroids, directed = FALSE,
     # This merges A->B and B->A into a single edge
     if(!directed) {
       swap <- graph$from > graph$to
+      if(verbose) cat("Oriented", sum(swap), "undirected edges\n")
       if(any(swap)) {
         tmp <- graph$from[swap]
         graph$from[swap] <- graph$to[swap]
@@ -1072,6 +1108,7 @@ contract_edges <- function(graph, nodes, clusters, centroids, directed = FALSE,
 
     # Include 'by' columns in grouping to prevent cross-mode consolidation
     g <- GRP(graph, c("from", "to", by))
+    if(verbose) cat("Contracting", fnrow(graph), "edges down to", g$N.groups, "edges\n")
     nam <- names(graph)
     res <- g$groups
     add_vars(res, pos = "front") <- list(edge = seq_row(res))
